"""The ``utils`` module contains tools used to perform common tasks in the audio
processing."""  # noqa: D205, D209

from enum import Enum

import numpy as np


class Criterion(Enum):
    """Represents a mathematical criterion.

    Some kinds of calculations can be done in different ways and often we want to have
    control over the way they are
    done. For example, when calculating the RMS of a signal, we can either take
    the maximum value of the RMS of each channel, or the mean value of the RMS of
    each channel.

    The following values are available:

    * ``MAX``: which value is the string ``'max'``.
    * ``MIN``: which value is the string ``'min'``.
    * ``MEAN``: which value is the string ``'mean'``.

    """

    MAX = "max"
    MIN = "min"
    MEAN = "mean"


def db_to_pcm(decibels: int, bit_depth: int) -> int:
    r"""Convert a dB value to a float value.

    The dB value is converted to a PCM value using the following formula:

    .. math::

        \text{PCM} = \left\lfloor 2^{bit\_depth - 1} \times 10^{\frac{decibels}{20}}
        \right\rceil

    where :math:`\lfloor x \rceil` is the *round* function, which approximates the
    result to the nearest integer (it introduces the quantization error).

    Parameters
    ----------
    decibels : int
        The dB value to convert.
    bit_depth : int
        The bit depth of the PCM value, this is used to calculate the maximum value of
        the PCM value.

    Returns
    -------
    int
        The PCM value corresponding to the given dB value.

    """
    return round(10 ** (decibels / 20) * 2 ** (bit_depth - 1))


def pcm_to_db(pcm: int | float, bit_depth: int) -> float:
    r"""Convert the given signal power to a dB value.

    The PCM value is converted to a dB value using the following formula:

    .. math::

        db = 20 \times log_{10}\left(\frac{\text{PCM}}{2^{bit\_depth - 1}}\right)

    Parameters
    ----------
    pcm : int
        The PCM value to convert.
    bit_depth : int
        The bit depth of the PCM value, this is used to calculate the maximum value of
        the PCM value.

    Returns
    -------
    float
        The dB value corresponding to the given PCM value.


    Notes
    -----
    Due to the formula the result on input ``0`` should be ``-inf``, nonetheless,
    as we are threating discrete quantities, we return the minimum value that can be
    represented by the given bit depth based on the following table:

    +---------+--------------------+
    | Bit     | Minimum value (dB) |
    +=========+====================+
    | 16      | -98                |
    +---------+--------------------+
    | 24      | -146               |
    +---------+--------------------+
    | 32      | -194               |
    +---------+--------------------+

    for a more detailed explanation see about `audio bit depth
    <https://en.wikipedia.org/wiki/Audio_bit_depth#Quantization>`_.

    """
    min_val = {
        "16": -98,
        "24": -146,
        "32": -194,
    }

    if pcm == 0:
        return min_val[str(bit_depth)]

    return 20 * np.log10(abs(pcm / 2 ** (bit_depth - 1)))


def rms(array: np.ndarray) -> float:
    r"""Calculate the RMS of the given array.

    The Root Mean Square (RMS) is the square root of the mean of the squares of the
    values in the array. It is a measure of the magnitude of a signal.
    It is calculated using the following formula:

    .. math::

        \text{RMS} = \sqrt{\frac{1}{n} \sum_{i=1}^{n} x_i^2}

    where :math:`x_i` is the value of the array at index :math:`i` and :math:`n` is the
    length of the array.

    Parameters
    ----------
    array : np.ndarray
        The array to calculate the RMS of.

    Returns
    -------
    float
        The RMS of the given array.

    """
    mean_squares = np.square(array).mean()
    if mean_squares == 0:
        return 0
    return np.sqrt(mean_squares)


def get_last_index(haystack: np.ndarray, threshold) -> int | tuple | None:
    """Get the index of the last occurrence of a value greater than ``threshold`` in the
    given array.

    Parameters
    ----------
    haystack : np.ndarray
        the array to inspect
    threshold : int
        the threshold to use

    Returns
    -------
    int | tuple | None
        The index of the last occurrence of a value greater than the given limit in the
        array, if the array is multi-dimensional it returns a tuple. If there is no
        match, None is returned

    """  # noqa: D205
    positions = np.where(haystack >= threshold)
    indexes = positions[0]
    if indexes.size == 0:
        return None
    if len(positions) > 1:
        return list(zip(*positions, strict=False))[-1]
    return indexes[-1]



import wave
from collections.abc import Callable, Iterable
from functools import lru_cache
from pathlib import Path
from typing import Annotated, Literal

import librosa
import numpy as np
from typing_extensions import Self, deprecated

from mpai_cae_arp.audio import utils
from mpai_cae_arp.audio._noise import Noise
from mpai_cae_arp.audio.utils import Criterion

Scale = Literal["db", "normal"]


class AudioWave:
    """A class to represent an audio wave.

    Attributes
    ----------
    bit : int
        the bit depth of the audio
    channels : int
        the number of channels of the audio
    samplerate : int
        the sample rate of the audio
    data : numpy.ndarray
        the data of the audio as a numpy array

    Raises
    ------
    ValueError
        if ``bit_depth`` is not 8, 16, 24 or 32
    ValueError
        if ``sample_rate`` is not between 8000 and 192000

    Examples
    --------
    Here is an example of how to create a new AudioWave object containing a sine wave
    with a frequency of 440 Hz, 44100 Hz sample rate, 16 bit depth and 1 channel:

    .. testcode::

        import numpy as np
        from mpai_cae_arp.audio import AudioWave

        freq = 440   # Hz
        sr = 44100   # sample rate
        bit = 16     # bit depth
        channels = 1 # number of channels

        # create an array of 1 second length
        time = np.arange(0, 1, 1/sr)

        # create a sine wave
        signal = np.sin(2 * np.pi * freq * time)

        # set the signal to the maximum value of the bit depth
        signal *= 2 ** (bit - 1)
        signal = np.int16(signal)

        # create the AudioWave object
        audio = AudioWave(signal, bit, channels, sr)

        print(audio)

    the code above will print the following output:

    .. testoutput::

        <AudioWave(bit: 16, chn: 1, sr: 44100)>

    """

    array: np.ndarray
    bit: int
    samplerate: int
    channels: int

    def __init__(
        self,
        data: np.ndarray,
        bit: int,
        channels: int,
        samplerate: Annotated[int, "Hz between 8000 and 192000"],
    ):
        if bit not in [8, 16, 24, 32]:
            msg = "bit must be 8, 16, 24 or 32"
            raise ValueError(msg)
        if samplerate not in range(8000, 192001):
            msg = "samplerate must be between 8 and 192 kHz"
            raise ValueError(msg)
        self.bit = bit
        self.channels = channels
        self.samplerate = samplerate
        self.array = data

    def __len__(self):
        return len(self.array)

    def __iter__(self):
        return iter(self.array)

    def __getitem__(self, key):
        return AudioWave(self.array[key], self.bit, self.channels, self.samplerate)

    def __eq__(self, __o: object) -> bool:
        if isinstance(__o, AudioWave):
            return (
                np.array_equal(self.array, __o.array)
                and self.bit == __o.bit
                and self.channels == __o.channels
                and self.samplerate == __o.samplerate
            )
        msg = f"cannot compare AudioWave with {type(__o)}"
        raise TypeError(msg)

    def __repr__(self) -> str:
        return (
            f"<AudioWave(bit: {self.bit}, chn: {self.channels}, sr: {self.samplerate})>"
        )

    @classmethod
    def from_file(
        cls,
        filepath: str,
        bit: int | None = None,
        channels: int | None = None,
        samplerate: int | None = None,
        bufferize: bool = False,
    ) -> Self:
        """Generate an AudioWave object from a file.

        It reads the binary headers of the file to automatically get the
        bit depth, the number of channels and the sample rate. If any of
        these values is given, it will be used instead of the one read
        from the file.

        """
        if bufferize:
            data = None
            for chunk in AudioWave.buffer_generator_from_file(filepath):
                if data is None:
                    bit = chunk.bit
                    channels = chunk.channels
                    samplerate = chunk.samplerate
                    data = chunk.array
                else:
                    data = np.concatenate((data, chunk.array), axis=0)

            if any(x is None for x in [data, bit, channels, samplerate]):
                msg = "Unable to read the file"
                raise ValueError(msg)

            return AudioWave(data, bit, channels, samplerate)  # type: ignore

        with wave.open(filepath, "rb") as filep:
            raw_data = filep.readframes(filep.getnframes())
            samplerate = filep.getframerate() if samplerate is None else samplerate
            channels = filep.getnchannels() if channels is None else channels
            bit = filep.getsampwidth() * 8 if bit is None else bit

        return AudioWave.from_bytes(raw_data, bit, channels, samplerate)

    @classmethod
    @lru_cache(maxsize=32)
    def read_file_metadata(cls, filepath: str) -> tuple[int, int, int]:
        """Read the metadata of an audio file.

        It reads the binary headers of the file to automatically get the
        bit depth, the number of channels and the sample rate.

        """
        with wave.open(filepath, "rb") as filep:
            samplerate = filep.getframerate()
            channels = filep.getnchannels()
            bit = filep.getsampwidth() * 8

        return bit, channels, samplerate

    @classmethod
    def buffer_generator_from_file(
        cls,
        filepath: str,
        buffer_size: int = 1024 * 1024 * 8,
    ):
        """Return a generator that yields AudioWave objects from a file.

        The generator will read the file in chunks of `buffer_size` bytes.

        Parameters
        ----------
        filepath: str
            The path to the file.
        buffer_size: int
            The size of audio chunks. Defaults to 1024*1024*8 (8 MegaBytes).

        Yields
        ------
        AudioWave
            An AudioWave object containing the audio data read from the file.

        """
        bit_dept, chn, sample_rate = cls.read_file_metadata(filepath)
        with wave.open(filepath, "rb") as wave_file:
            frames: bytes = wave_file.readframes(buffer_size)
            while frames != b"":
                yield cls.from_bytes(frames, bit_dept, chn, sample_rate)
                frames = wave_file.readframes(buffer_size)

    @classmethod
    def from_bytes(
        cls,
        raw_data: bytes,
        bit: int,
        channels: int,
        samplerate: int,
    ) -> Self:
        r"""Generate an AudioWave object from a stream of bytes.

        It is assumed that the data is in little endian format and that the bytes
        are signed integers.

        Parameters
        ----------
        raw_data: bytes
            The stream of bytes.
        bit: int
            The bit depth of the audio.
        channels: int
            The number of channels of the audio.
        samplerate: int
            The sample rate of the audio.

        Example
        -------

        The following example shows how to create an AudioWave object from a stream of
        bytes.

        .. doctest::

            >>> import numpy as np
            >>> from mpai_cae_arp.audio import AudioWave
            >>> audio = AudioWave.from_bytes(b'\x00\x00\x00\x00', 16, 2, 44100)
            >>> np.array_equal(audio.array, np.array([[0, 0]]))
            True

        """
        data = np.array(
            [
                int.from_bytes(
                    raw_data[i : i + bit // 8],
                    byteorder="little",
                    signed=True,
                )
                for i in range(0, len(raw_data), bit // 8)
            ],
            dtype=np.int64,
        )  # int32 (default in Windows) leds to overflow
        data = np.reshape(data, (-1, channels))
        return cls(data, bit, channels, samplerate)

    def save(self, filepath: str, force: bool = False) -> None:
        """Save the audio as a wave file at the given path.

        Parameters
        ----------
        filepath: str
            The path where to save the audio.
        force: bool
            If True, it will create the directory if it does not exist.

        Raises
        ------
        ValueError
            if the directory does not exist and force is False

        Examples
        --------
        .. code-block:: python

            from mpai_cae_arp.audio import AudioWave

            audio = AudioWave(np.array([[0, 0]]), 16, 2, 44100)
            audio.save('wrong/path/test.wav') # raises ValueError
            audio.save('force/path/test.wav', force=True) # creates the path and saves

        """
        if not Path(filepath).parent.exists():
            if force:
                Path(filepath).parent.mkdir(parents=True, exist_ok=True)
            else:
                msg = f"Directory {Path(filepath).parent} does not exist"
                raise ValueError(msg)

        with wave.open(filepath, "wb") as filep:
            filep.setframerate(self.samplerate)
            filep.setnchannels(self.channels)
            filep.setsampwidth(self.bit // 8)
            filep.setnframes(self.number_of_frames())
            filep.writeframesraw(self.get_raw())

    def set_sample_rate(self, samplerate: int) -> None:
        """Set the sample rate of the audio.

        .. versionadded:: 0.4.0

        Parameters
        ----------
        samplerate: int
            The new sample rate.

        """
        self.array = librosa.resample(
            self.array,
            orig_sr=self.samplerate,
            target_sr=samplerate,
        )
        self.samplerate = samplerate

    def get_raw(self) -> bytes:
        """Get the raw data of the audio.

        Returns
        -------
        bytes
            A bytes stream in the form (left right left right ...), where left and right
            are the samples for the left and right channels respectively for each frame
            in little endian format and signed.

        """
        data: np.ndarray = np.reshape(self.array, (-1,))
        bytes_array = b"".join(
            [
                int(x).to_bytes(self.bit // 8, byteorder="little", signed=True)
                for x in data
            ],
        )

        return bytes_array

    def number_of_frames(self) -> int:
        """Get the number of frames in the audio.

        A frame is a sample for each channel. In a stereo audio, a frame is composed by
        two samples, one for each channel, in the form [left, right], where left and
        right are the samples for the left and right channels respectively.

        Returns
        -------
        int
            The number of frames in the audio

        """
        return len(self.array)

    def duration(self) -> float:
        r"""Get the duration of the audio in seconds.

        The duration is calculated by dividing the number of frames by the sample rate.

        .. math::

            time = \frac{frames}{sample\_rate}

        Returns
        -------
        float
            the duration of the audio in seconds

        """
        return self.number_of_frames() / self.samplerate

    def get_mfcc(self, n_mfcc: int = 13) -> np.ndarray:
        """Get the Mel-frequency cepstral coefficients (MFCC) of the audio.

        Parameters
        ----------
        n_mfcc: int
            The number of MFCC to return. Default is 13.

        Returns
        -------
        np.ndarray
            The mean of the first ``n_mfcc`` mfcc over the entire audio signal.
            If the audio is multichannel, the mean of the mfcc of each channel is
            returned.

        """
        mfcc_per_ch = []

        for channel in range(self.channels):
            signal = self.get_channel(channel).array
            signal = signal / (2 ^ (self.bit - 1))  # normalize the signal
            mfccs = librosa.feature.mfcc(y=signal, sr=self.samplerate, n_mfcc=n_mfcc)
            mean_mfccs = []
            for elem in mfccs:
                mean_mfccs.append(np.mean(elem))
            mfcc_per_ch.append(mean_mfccs)

        # return the mean of the mfcc of each channel
        return np.mean(mfcc_per_ch, axis=0)

    def rms(self, mode: Criterion = Criterion.MAX, scale: Scale = "normal") -> float:
        r"""Get the root-mean-square of the audio.

        The root-mean-square of the audio. It measures the power level of the
        entire signal. It is calculated by taking the square root of the mean of
        the square of the samples. In multichannel audio, the rms is calculated by
        taking the mean or the max of the rms of each channel.

        Parameters
        ----------
        mode: Criterion
            The mode to use to calculate the rms.
            It can be ``Criterion.MEAN`` or ``Criterion.MAX``.
            If ``Criterion.MEAN``, the rms is calculated by taking the mean of the rms
            of each channel. If ``Criterion.MAX``, the rms is calculated by taking the
            maximum of the rms between all channels. Default is ``Criterion.MAX``.
        scale: Scale
            The scale to use to calculate the rms. It can be ``"db"`` or ``"normal"``.

        Raises
        ------
        ValueError
            if the mode is not ``MEAN`` or ``MAX``

        Returns
        -------
        float
            The rms of the audio.

        """
        functions: dict[Scale, Callable[[int | float, int], float]] = {
            "db": utils.pcm_to_db,
            "normal": lambda x, _: x,
        }
        try:
            scale_funciton = functions[scale]
        except KeyError as err:
            msg = "Invalid scale"
            raise ValueError(msg) from err

        if self.channels == 1:
            return scale_funciton(utils.rms(self.array), self.bit)
        channels_rms = []
        for channel in range(self.channels):
            channels_rms.append(utils.rms(self.array[:, channel]))

        match mode:
            case Criterion.MEAN:
                return scale_funciton(np.array(channels_rms).mean(), self.bit)
            case Criterion.MAX:
                return scale_funciton(max(channels_rms), self.bit)
            case _:
                msg = "Invalid criterion"
                raise ValueError(msg)

    @deprecated("Use AudioWave.rms(scale='db') instead.")
    def db_rms(self, mode: Criterion = Criterion.MAX) -> float:
        """Get the rms of audio in decibel scale.

        It is a wrapper of ``self.rms()``, that converts the returned value to decibels.

        .. deprecated:: 0.5.2
            Use :meth:`AudioWave.rms` with argument :code:`scale="db"` instead.

        .. seealso::
            see also the method :meth:`AudioWave.rms` for a more detailed description
            of the parameters.

        """
        return utils.pcm_to_db(self.rms(mode), self.bit)

    def get_channel(self, channel: int) -> Self:
        """Create a new AudioWave object containing only the given channel.

        Parameters
        ----------
        channel: int
            the channel to extract from the audio, where 0 is the first channel,
            1 the second and so on.

        Raises
        ------
        IndexError
            if the channel is not found.

        Returns
        -------
        AudioWave
            An AudioWave object containing only the given channel.

        Example
        -------

        .. code-block:: python

            from mpai_cae_arp.audio import AudioWave

            # let's say 'sample.wav' is a stereo audio

            # returns an AudioWave object containing only the first channel
            audio_left = AudioWave.from_file('sample.wav').get_channel(0)

            # returns an AudioWave object containing only the second channel
            audio_right = AudioWave.from_file('sample.wav').get_channel(1)

            # raises IndexError because there are only two channels
            audio_third = AudioWave.from_file('sample.wav').get_channel(2)

        """
        if channel not in range(self.channels):
            msg = "Channel not found"
            raise IndexError(msg)

        if self.channels == 1:
            return AudioWave(self.array.squeeze(), self.bit, 1, self.samplerate)
        return AudioWave(self.array[:, channel], self.bit, 1, self.samplerate)

    def get_silence_slices(
        self,
        noise_list: Iterable[Noise],
        length: int,
    ) -> dict[str, list[tuple[int, int]]]:
        """Get the slices of silence in the given signal.

        Passing a list of ``Noise`` instances to the ``noise_list`` parameter,
        the function will return a dictionary of slices of silence in the given signal.
        The dictionary will have the label of the noise as key and a list of tuples as
        value. Each tuple will contain the starting and ending frames of a slice of
        silence.

        Parameters
        ----------
        noise_list: Iterable[Noise]
            the noise bands to use to detect the silence
        length: int
            the length of a slice of silence in milliseconds

        Raises
        ------
        ValueError
            if ``length`` is less than 1

        Returns
        -------
        dict[str, list[tuple[int, int]]]
            A dictionary with the label of the noise as key and a list of tuples as
            value. Each tuple will contain the starting and ending frames of a slice of
            silence. If the signal is completely silent, or the ``min_length`` required
            is greater than the signal length, the dictionary will contain an empty list
            for each noise in the ``noise_list``.

        Example
        -------

        For example, if we have the following audio:

        .. doctest::

            >>> from mpai_cae_arp.audio import AudioWave, Noise
            >>> noise_list = [Noise("noise1", -10, -20), Noise("noise2", -30, -40)]
            >>> array = np.array([10000 for _ in range(8000)])
            >>> audio = AudioWave(array, 16, 1, 8000)
            >>> audio.get_silence_slices(noise_list, 500)
            {'noise1': [(0, 4000), (4000, 8000)], 'noise2': []}


        .. note::
            The ``AudioWave`` abstraction makes possible to analyze the signal at a
            frame level. Nonetheless, the function will scan the signal at a millisecond
            interval.
            This is because in a millisecond there are many frames that can have a rms
            value which belongs to different ``Noise`` instances, but those variations
            are not relevant for the purpose of detecting silence since are not
            perceptible by the human ear.

        """
        # sanity checks on the parameters
        if not isinstance(length, int) or length < 1:
            msg = "min_length must be an integer greater or equal 1"
            raise ValueError(msg)

        window_frames = length * self.samplerate // 1000
        last_frame = self.number_of_frames() - window_frames

        # create an empty array of indexes for each noise to filter
        idxs = {noise.label: [] for noise in noise_list}

        # if the signal is too short, return an empty dict
        if last_frame < 1 or noise_list == []:
            return idxs

        # find the thresholds to be used to detect the noise
        upper_noise_limit = max(noise.db_max for noise in noise_list)
        lower_noise_limit = min(noise.db_min for noise in noise_list)

        i = 0

        while i <= last_frame:
            chunk = self[i : i + window_frames]
            max_pos = utils.get_last_index(
                chunk.array,
                utils.db_to_pcm(upper_noise_limit, chunk.bit),
            )
            # if there is a value over the threshold go to its index and start
            # seeking from here
            if max_pos is not None:
                # if multidimensional array get only the x value
                if isinstance(max_pos, tuple):
                    max_pos = max_pos[0]
                i += max_pos + 1
            # else all the signal is under the threshold
            else:
                # classification of the chunk based on rms
                for noise in noise_list:
                    if noise.db_min <= chunk.db_rms() < noise.db_max:
                        idxs[noise.label].append((i, i + window_frames))
                        break
                # if the signal power is under the minimum, assign it to the quieter
                # noise
                if chunk.db_rms() < lower_noise_limit:
                    idxs[min(noise_list).label].append((i, i + window_frames))
                i += window_frames

        return idxs



from abc import abstractmethod

from grpc import Channel, Server, ServicerContext

from .arp_pb2 import InfoRequest, InfoResponse, JobRequest, JobResponse

class AIMStub:
    def __init__(self, channel: Channel) -> None: ...

class AIMServicer:
    @abstractmethod
    def getInfo(
        self,
        request: InfoRequest,
        context: ServicerContext,
    ) -> InfoResponse: ...
    @abstractmethod
    def work(self, request: JobRequest, context: ServicerContext) -> JobResponse: ...

class AIM:
    @staticmethod
    def getInfo(
        request: InfoRequest,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure: bool = False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ) -> InfoResponse: ...
    @staticmethod
    def work(
        request: JobRequest,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure: bool = False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ) -> JobResponse: ...

def add_AIMServicer_to_server(servicer: AIMServicer, server: Server) -> None: ...



"""The types module contains the data types specified in the MPAI-CAE-ARP standard.

Here we have the following data types:

- :class:`Irregularity`: an event of interest that has been detected in the audio or
  video signal.
- :class:`IrregularityFile`: a file containing a list of irregularities that is used to
   exchange messages between the different components of the system.
- :class:`IrregularityProperties`: a set of properties that describe an irregularity
   (only for audio ones).
- :class:`IrregularityType`: the type of irregularity.
- :class:`Restoration`: a restoration that has been applied to the audio signal.
- :class:`EditingList`: a list of restorations.

"""

from . import schema
from ._irregularity import (
    Irregularity,
    IrregularityFile,
    IrregularityProperties,
    IrregularityType,
    Source,
)
from ._restoration import EditingList, Restoration

__all__ = [
    "Irregularity",
    "IrregularityFile",
    "IrregularityProperties",
    "IrregularityType",
    "Restoration",
    "EditingList",
    "schema",
    "Source",
]



import uuid

from pydantic import BaseModel, Field
from typing_extensions import Self

from mpai_cae_arp.audio.standards import EqualizationStandard, SpeedStandard
from mpai_cae_arp.files import File, FileType


class Restoration(BaseModel):
    """
    .. versionadded:: 0.4.0.

    .. todo:: Add a description
    """

    id: uuid.UUID = Field(
        default_factory=uuid.uuid4,
        serialization_alias="RestorationID",
    )
    preservation_audio_file_start: str = Field(
        serialization_alias="PreservationAudioFileStart",
    )
    preservation_audio_file_end: str = Field(
        serialization_alias="PreservationAudioFileEnd",
    )
    restored_audio_file_URI: str = Field(serialization_alias="RestoredAudioFileURI")
    reading_backwards: bool = Field(serialization_alias="ReadingBackwards")
    applied_speed_standard: SpeedStandard = Field(
        serialization_alias="AppliedSpeedStandard",
    )
    applied_sample_frequency: int = Field(serialization_alias="AppliedSampleFrequency")
    applied_equalization_standard: EqualizationStandard = Field(
        serialization_alias="AppliedEqualisationStandard",
    )


class EditingList(BaseModel):
    """
    .. versionadded:: 0.4.0.

    .. todo:: Add a description
    """

    original_speed_standard: SpeedStandard = Field(
        serialization_alias="OriginalSpeedStandard",
    )
    original_equalization_standard: EqualizationStandard = Field(
        serialization_alias="OriginalEqualisationStandard",
    )
    original_sample_frequency: int = Field(
        serialization_alias="OriginalSampleFrequency",
    )
    restorations: list[Restoration] = Field(serialization_alias="Restorations")

    def add(self, restoration: Restoration) -> Self:
        self.restorations.append(restoration)
        return self

    def remove(self, restoration: Restoration) -> Self:
        self.restorations.remove(restoration)
        return self

    def remove_by_id(self, restoration_id: uuid.UUID) -> Self:
        filtered = list(filter(lambda r: r.id != restoration_id, self.restorations))

        if len(filtered) == len(self.restorations):
            msg = f"Restoration with ID {restoration_id} not found."
            raise ValueError(msg)

        self.restorations = filtered
        return self

    def save_as_json_file(self, path: str) -> None:
        File(path=path, format=FileType.JSON).write_content(
            self.model_dump(mode="json", by_alias=True),  # type: ignore
        )



"""Handle files operations."""

import json
from enum import Enum
from io import TextIOWrapper
from pathlib import Path
from typing import Literal

import yaml
from pydantic import BaseModel


class FileAction(str, Enum):
    """The action to perform on the file at opening.

    It can be one of the following:

    - ``r``: read
    - ``w``: write
    - ``a``: append

    """

    READ = "r"
    WRITE = "w"
    APPEND = "a"


class FileType(str, Enum):
    """The file type.

    It can be one of the following:

    - ``yaml``
    - ``json``

    """

    YAML = "yaml"
    JSON = "json"


Encoding = Literal["utf-8"]


class File(BaseModel):
    """Class representing a file, with its path, encoding and format."""

    encoding: Encoding = "utf-8"
    format: FileType  # noqa: A003
    path: str

    def open(self, action: FileAction) -> TextIOWrapper:  # noqa: A003
        """Open the file with the given action.

        Parameters
        ----------
        action : FileAction
            the action to perform on the file at opening

        Returns
        -------
        TextIOWrapper
            the file descriptor

        """
        return Path(self.path).open(  # noqa: SIM115
            mode=action.value,
            encoding=self.encoding,
        )

    def get_content(self) -> dict:
        """Return the file content.

        Raises
        ------
        ValueError
            if the format is not supported

        Returns
        -------
        dict
            the parsed content of the file

        """
        with self.open(FileAction.READ) as filep:
            if self.format == FileType.YAML:
                content = yaml.safe_load(filep)
                return content
            if self.format == FileType.JSON:
                content = json.load(filep)
            else:
                msg = "Format not supported"
                raise ValueError(msg)

        return content

    def write_content(self, content: dict | str) -> None:
        """Write the given content in the file.

        Parameters
        ----------
        content : dict | str
            the content to write in the file
            if content is a string, it will be parsed as a JSON string

        """
        if isinstance(content, str):
            try:
                content = json.loads(json.loads(content))
            except json.decoder.JSONDecodeError as exc:
                msg = "Content str is not a valid JSON string"
                raise ValueError(msg) from exc

        with self.open(FileAction.WRITE) as filep:
            if self.format == FileType.YAML:
                yaml.safe_dump(content, filep)
            elif self.format == FileType.JSON:
                json.dump(content, filep, indent=4)
            else:
                msg = "Format not supported"
                raise ValueError(msg)



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from . import arp_pb2 as arp__pb2


class AIMStub:
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel: grpc.Channel):
        """Create a stub for the AIM service.

        Parameters
        ----------
        channel:
            A grpc.Channel.

        """
        self.getInfo = channel.unary_unary(
            "/arp.AIM/getInfo",
            request_serializer=arp__pb2.InfoRequest.SerializeToString,
            response_deserializer=arp__pb2.InfoResponse.FromString,
        )
        self.work = channel.unary_stream(
            "/arp.AIM/work",
            request_serializer=arp__pb2.JobRequest.SerializeToString,
            response_deserializer=arp__pb2.JobResponse.FromString,
        )


class AIMServicer:
    """Missing associated documentation comment in .proto file."""

    def getInfo(self, request, context):
        """Get information about the AIM."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        msg = "Method not implemented!"
        raise NotImplementedError(msg)

    def work(self, request, context):
        """Do the AIM task."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        msg = "Method not implemented!"
        raise NotImplementedError(msg)


def add_AIMServicer_to_server(servicer, server):
    """Add a servicer to a server."""
    rpc_method_handlers = {
        "getInfo": grpc.unary_unary_rpc_method_handler(
            servicer.getInfo,
            request_deserializer=arp__pb2.InfoRequest.FromString,
            response_serializer=arp__pb2.InfoResponse.SerializeToString,
        ),
        "work": grpc.unary_stream_rpc_method_handler(
            servicer.work,
            request_deserializer=arp__pb2.JobRequest.FromString,
            response_serializer=arp__pb2.JobResponse.SerializeToString,
        ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
        "arp.AIM",
        rpc_method_handlers,
    )
    server.add_generic_rpc_handlers((generic_handler,))


# This class is part of an EXPERIMENTAL API.
class AIM:
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def getInfo(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        """Exposes the getInfo RPC.

        EXPERIMENTAL.

        """
        return grpc.experimental.unary_unary(
            request,
            target,
            "/arp.AIM/getInfo",
            arp__pb2.InfoRequest.SerializeToString,
            arp__pb2.InfoResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def work(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        """Exposes the work RPC.

        EXPERIMENTAL.

        """
        return grpc.experimental.unary_stream(
            request,
            target,
            "/arp.AIM/work",
            arp__pb2.JobRequest.SerializeToString,
            arp__pb2.JobResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )



"""Utilities for the network module."""
from concurrent import futures

from grpc import Server, server
from typing_extensions import Self

from .arp_pb2_grpc import AIMServicer, add_AIMServicer_to_server


class ServerBuilder:
    """Builder for the server."""

    _server: Server

    def __init__(self) -> None:
        self._server = server(futures.ThreadPoolExecutor(max_workers=8))

    def set_port(self, port: int | str = 50051) -> Self:
        """Set the port of the server.

        Parameters
        ----------
        port : int | str, optional
            the port of the server, by default 50051

        """
        self._server.add_insecure_port(f"[::]:{port}")
        return self

    def add_servicer(self, servicer: AIMServicer) -> Self:
        """Add a servicer to the server.

        A servicer is a class that implements the AIMServicer interface.

        See Also
        --------
        AIMServicer

        """
        add_AIMServicer_to_server(servicer, self._server)
        return self

    def build(self) -> Server:
        """Get the built server."""
        return self._server

    def new(self) -> Self:
        """Reset the builder to its initial state."""
        return ServerBuilder()



from typing_extensions import Self

class InfoRequest:
    def __init__(self) -> None: ...
    @classmethod
    def FromString(cls, s: str) -> Self: ...
    @classmethod
    def SerializeToString(cls, s: Self) -> str: ...
    ...

class JobRequest:
    working_dir: str
    files_name: str
    index: int
    def __init__(self, working_dir: str, files_name: str, index: int) -> None: ...
    @classmethod
    def FromString(cls, s: str) -> Self: ...
    @classmethod
    def SerializeToString(cls, s: Self) -> str: ...
    ...

class JobResponse:
    status: str
    message: str
    def __init__(self, status: str, message: str) -> None: ...
    @classmethod
    def FromString(cls, s: str) -> Self: ...
    @classmethod
    def SerializeToString(cls, s: Self) -> str: ...
    ...

class Contact:
    name: str
    email: str
    def __init__(self, name: str, email: str) -> None: ...
    @classmethod
    def FromString(cls, s: str) -> Self: ...
    @classmethod
    def SerializeToString(cls, s: Self) -> str: ...
    ...

class License:
    name: str
    url: str
    def __init__(self, name: str, url: str) -> None: ...
    @classmethod
    def FromString(cls, s: str) -> Self: ...
    @classmethod
    def SerializeToString(cls, s: Self) -> str: ...
    ...

class InfoResponse:
    title: str
    description: str
    version: str
    contact: Contact
    license: License
    def __init__(
        self,
        title: str,
        description: str,
        version: str,
        contact: Contact,
        license: License,
    ) -> None: ...
    @classmethod
    def FromString(cls, s: str) -> Self: ...
    @classmethod
    def SerializeToString(cls, s: Self) -> str: ...
    ...



from dataclasses import dataclass

CMP_ERROR = "Cannot compare Noise with not Noise"


@dataclass
class Noise:
    """A class to represent a noise.

    A noise is a signal that resides in a certain range of decibels, like a sort of
    filter.

    Attributes
    ----------
    label : str
        a label which describes this noise
    db_max : int
        the maximum value of this noise in decibels (must be less than or equal to 0)
    db_min : int
        the minimum value of this noise in decibels (must be less than or equal to 0)

    Raises
    ------
    ValueError
        if ``db_max`` is less than ``db_min``, or if ``db_min`` or ``db_max`` are
        greater than 0


    .. topic:: More about noise classes

        Noise classes of interest are defined in :cite:`10.1162/comj_a_00487`, where we
        can find three main notable classes:

        1. type A, from :math:`-50dB` to :math:`-63dB`, which is noise in the middle of
           a recording, i.e., silence between spoken words;
        2. type B, from :math:`-63dB` to :math:`-69dB`, which is noise of the recording
           head without any specific input signal;
        3. type C, from :math:`-69dB` to :math:`-72dB`, which is noise coming from
           sections of pristine tape.

    """

    _label: str
    _db_max: int
    _db_min: int

    def __init__(self, label: str, db_max: int, db_min: int) -> None:
        if db_max < db_min:
            msg = "db_max must be greater than db_min"
            raise ValueError(msg)
        if db_min > 0 or db_max > 0:
            msg = "db_min and db_max must be negative or zero"
            raise ValueError(msg)

        self._label = label
        self._db_max = db_max
        self._db_min = db_min

    @property
    def label(self) -> str:
        return self._label

    @property
    def db_range(self) -> int:
        return self.db_max - self.db_min

    @property
    def db_max(self) -> int:
        return self._db_max

    @property
    def db_min(self) -> int:
        return self._db_min

    def __lt__(self, __o: object) -> bool:
        if not isinstance(__o, Noise):
            raise TypeError(CMP_ERROR)
        return self.db_min < __o.db_min

    def __gt__(self, __o: object) -> bool:
        if not isinstance(__o, Noise):
            raise TypeError(CMP_ERROR)
        return self.db_max > __o.db_max

    def __le__(self, __o: object) -> bool:
        if not isinstance(__o, Noise):
            raise TypeError(CMP_ERROR)
        return self < __o or self == __o

    def __repr__(self) -> str:
        return f"<Noise({self.label}, {self.db_min}dB, {self.db_max}dB)>"



import uuid
from enum import Enum
from typing import Any

from pydantic import BaseModel, Field
from typing_extensions import Self, deprecated

from mpai_cae_arp.audio.standards import EqualizationStandard, SpeedStandard
from mpai_cae_arp.files import File, FileType
from mpai_cae_arp.time import time_to_seconds


class IrregularityType(str, Enum):
    BRANDS_ON_TAPE = "b"
    SPLICE = "sp"
    START_OF_TAPE = "sot"
    ENDS_OF_TAPE = "eot"
    DAMAGED_TAPE = "da"
    DIRT = "di"
    MARKS = "m"
    SHADOWS = "s"
    WOW_AND_FLUTTER = "wf"
    PLAY_PAUSE_STOP = "pps"
    SPEED = "ssv"
    EQUALIZATION = "esv"
    SPEED_AND_EQUALIZATION = "ssv"  # noqa: PIE796
    BACKWARD = "sb"


class Source(str, Enum):
    AUDIO = "a"
    VIDEO = "v"
    BOTH = "b"


Json = dict[str, Any]


class IrregularityProperties(BaseModel):
    reading_speed: SpeedStandard = Field(serialization_alias="ReadingSpeedStandard")
    reading_equalisation: EqualizationStandard = Field(
        serialization_alias="ReadingEqualisationStandard",
    )
    writing_speed: SpeedStandard = Field(serialization_alias="WritingSpeedStandard")
    writing_equalisation: EqualizationStandard = Field(
        serialization_alias="WritingEqualisationStandard",
    )

    @classmethod
    @deprecated("Use directly the constructor instead.")
    def from_json(cls, json_property: Json) -> Self:
        return cls(
            reading_speed=SpeedStandard(json_property["ReadingSpeedStandard"]),
            reading_equalisation=EqualizationStandard(
                json_property["ReadingEqualisationStandard"],
            ),
            writing_speed=SpeedStandard(json_property["WritingSpeedStandard"]),
            writing_equalisation=EqualizationStandard(
                json_property["WritingEqualisationStandard"],
            ),
        )

    @deprecated("Use IrregularityProperties.json instead.")
    def to_json(self) -> Json:
        """Convert the properties to a dictionary.

        .. deprecated:: 0.4.0
            Use :meth:`IrregularityProperties.json` instead.

        """
        return {
            "ReadingSpeedStandard": self.reading_speed.value,
            "ReadingEqualisationStandard": self.reading_equalisation.value,
            "WritingSpeedStandard": self.writing_speed.value,
            "WritingEqualisationStandard": self.writing_equalisation.value,
        }

    def get_irregularity_type(self) -> IrregularityType | None:
        if self.reading_equalisation != self.writing_equalisation:
            if self.reading_speed != self.writing_speed:
                return IrregularityType.SPEED_AND_EQUALIZATION
            return IrregularityType.EQUALIZATION
        if self.reading_speed != self.writing_speed:
            return IrregularityType.SPEED
        return None


class Irregularity(BaseModel):
    id: uuid.UUID = Field(
        default_factory=uuid.uuid4,
        serialization_alias="IrregularityID",
    )
    source: Source = Field(serialization_alias="Source")
    time_label: str = Field(serialization_alias="TimeLabel")
    type: IrregularityType | None = Field(
        default=None,
        serialization_alias="IrregularityType",
    )
    properties: IrregularityProperties | None = Field(
        default=None,
        serialization_alias="IrregularityProperties",
    )
    image_URI: str | None = Field(default=None, serialization_alias="ImageURI")
    audio_block_URI: str | None = Field(
        default=None,
        serialization_alias="AudioBlockURI",
    )

    @classmethod
    @deprecated("Use directly the constructor instead.")
    def from_json(cls, json_irreg: Json) -> Self:
        """Convert a JSON irregularity to an Irregularity object.

        .. deprecated:: 0.5.3
            Use :class:`Irregularity` constructor instead.

        """
        properties: IrregularityProperties | None = json_irreg.get(
            "IrregularityProperties",
        )
        irregularity_type: IrregularityType | None = None

        if properties is not None:
            properties = IrregularityProperties.from_json(
                json_irreg["IrregularityProperties"],
            )

        raw_irreg_type: str | None = json_irreg.get("IrregularityType")

        if raw_irreg_type:
            if raw_irreg_type is not (
                IrregularityType.SPEED.value
                or IrregularityType.SPEED_AND_EQUALIZATION.value
            ):
                irregularity_type = IrregularityType(raw_irreg_type)
            else:
                if properties is None:
                    msg = "Unexpected value: properties is None"
                    raise ValueError(msg)
                if properties.reading_equalisation != properties.writing_equalisation:
                    irregularity_type = IrregularityType.SPEED_AND_EQUALIZATION
                else:
                    irregularity_type = IrregularityType.SPEED

        return cls(
            id=uuid.UUID(json_irreg["IrregularityID"]),
            source=Source(json_irreg["Source"]),
            time_label=json_irreg["TimeLabel"],
            type=irregularity_type,
            properties=properties,
            image_URI=json_irreg.get("ImageURI"),
            audio_block_URI=json_irreg.get("AudioBlockURI"),
        )

    @deprecated("Use Irregularity.json instead.")
    def to_json(self) -> Json:
        """Create a dictionary with the irregularity information.

        .. deprecated:: 0.4.0
            Use :func:`Irregularity.json` instead.

        """
        dictionary: Json = {
            "IrregularityID": str(self.id),
            "Source": self.source.value,
            "TimeLabel": self.time_label,
        }

        if self.type:
            dictionary["IrregularityType"] = self.type.value

        if self.image_URI:
            dictionary["ImageURI"] = self.image_URI

        if self.audio_block_URI:
            dictionary["AudioBlockURI"] = self.audio_block_URI

        if self.properties:
            dictionary["IrregularityProperties"] = self.properties.to_json()

        return dictionary


class IrregularityFile(BaseModel):
    irregularities: list[Irregularity] = Field(serialization_alias="Irregularities")
    offset: int | None = Field(default=None, serialization_alias="Offset")

    def __eq__(self, __o: object) -> bool:
        if not isinstance(__o, IrregularityFile):
            return False

        return self.irregularities == __o.irregularities and self.offset == __o.offset

    @classmethod
    @deprecated("Use directly the constructor instead.")
    def from_json(cls, json_irreg: Json) -> Self:
        """Create an IrregularityFile object from a JSON dictionary.

        .. deprecated:: 0.4.0
            Use :class:`IrregularityFile` constructor instead.

        """
        irregularities: list[Irregularity] = []

        for irreg in json_irreg["Irregularities"]:
            irregularities.append(Irregularity.from_json(irreg))

        return IrregularityFile(
            irregularities=irregularities,
            offset=json_irreg.get("Offset"),
        )

    @deprecated("Use IrregularityFile.json instead.")
    def to_json(self) -> Json:
        """Convert the IrregularityFile to a dictionary.

        .. deprecated:: 0.4.0
            Use :meth:`IrregularityFile.json` instead.

        """
        dictionary: Json = {
            "Irregularities": [
                irregularity.to_json() for irregularity in self.irregularities
            ],
        }

        if self.offset:
            dictionary["Offset"] = self.offset

        return dictionary

    def add(self, irregularity: Irregularity) -> Self:
        """Add an irregularity to the list of irregularities.

        Parameters
        ----------
        irregularity : Irregularity
            the irregularity to add

        Raises
        ------
        TypeError
            if the irregularity is not a py:class:`Irregularity` object

        """
        if not isinstance(irregularity, Irregularity):
            msg = "IrregularityFile.add() expects an Irregularity object"
            raise TypeError(msg)
        self.irregularities.append(irregularity)
        self.order()
        return self

    def order(self) -> Self:
        """Sort the irregularities by time label."""
        self.irregularities.sort(key=lambda x: time_to_seconds(x.time_label))
        return self

    def join(self, other) -> Self:
        """Merge two irregularity files."""
        if not isinstance(other, IrregularityFile):
            msg = "other must be an instance of IrregularityFile"
            raise TypeError(msg)
        self.irregularities += other.irregularities
        self.order()
        return self

    def save_as_json_file(self, path: str) -> None:
        """Save the irregularity file as a JSON file at the given path.

        .. versionadded:: 0.4.0

        """
        File(path=path, format=FileType.JSON).write_content(
            self.model_dump(by_alias=True),  # type: ignore
        )



"""
.. deprecated:: 0.2.0.

    Use rich library instead.

This module contains functions to print and format text in the console.

"""
import warnings
from collections.abc import Iterable
from enum import Enum

from typing_extensions import deprecated

warnings.warn(
    "the mpai_cae_arp.io module is deprecated",
    DeprecationWarning,
    stacklevel=2,
)

END = "\033[0m"


@deprecated("Use rich library instead")
class Style(Enum):
    """Styles for console output."""

    BOLD = "\033[1m"
    UNDERLINE = "\033[4m"


@deprecated("Use rich library instead")
class Color(Enum):
    """Colors for console output."""

    PURPLE = "\033[95m"
    CYAN = "\033[96m"
    DARK_CYAN = "\033[36m"
    BLUE = "\033[94m"
    GREEN = "\033[92m"
    YELLOW = "\033[93m"
    RED = "\033[91m"


@deprecated("Use rich library instead")
def prettify(
    text: str,
    color: Color | None = None,
    styles: Iterable[Style] | None = None,
) -> str:
    """Format a string with some styles.

    Parameters
    ----------
    text : str
        string to format
    color : Color, optional
        color to use
    styles : list[Style], optional
        styles to use

    Returns
    -------
    str
        formatted string

    """
    prepend = ""

    if color:
        prepend += color.value
    if styles:
        for style in styles:
            prepend += style.value

    return prepend + text + END


@deprecated("Use rich library instead")
def pprint(
    text: str,
    color: Color | None = None,
    styles: list[Style] | None = None,
) -> None:
    """Format a string with some styles and prints it.

    Parameters
    ----------
    text : str
        string to format
    color : Color, optional
        color to use
    styles : list[Style], optional
        styles to use

    """
    print(prettify(text, color, styles))



"""Classes for the API informations."""

from pydantic import BaseModel, ConfigDict, Field


class Contact(BaseModel):
    """Contact information for the owners of the API.

    Attributes
    ----------
    name : str
        Name of the contact person.
    email : str
        Email of the contact person.

    """

    name: str = Field(..., description="Name of the contact person.")
    email: str = Field(..., description="Email of the contact person.")
    model_config = ConfigDict(  # type: ignore
        json_schema_extra={"example": {"name": "John Doe", "email": "email@email.com"}},
    )


class License(BaseModel):
    """License information for the API.

    Attributes
    ----------
    name : str
        Name of the license.
    url : str
        URL of the license.

    """

    name: str = Field(..., description="Name of the license.")
    url: str = Field(..., description="URL of the license.")
    model_config = ConfigDict(
        json_schema_extra={  # type: ignore
            "example": {"name": "MIT", "url": "https://opensource.org/licenses/MIT"},
        },
    )


class Info(BaseModel):
    """Information about the API.

    Attributes
    ----------
    title : str
        The title of the API.
    description : str
        A short description of the API.
    version : str
        The version of the API.
    contact : Contact
        Contact information for the owners of the API.
    license_info : License
        License information for the API.

    """

    title: str = Field(..., description="The title of the API.")
    description: str = Field(..., description="A short description of the API.")
    version: str = Field(..., description="The version of the API.")
    contact: Contact = Field(
        ...,
        description="Contact information for the owners of the API.",
    )
    license_info: License = Field(..., description="License information for the API.")
    model_config = ConfigDict(  # type: ignore
        json_schema_extra={
            "example": {
                "title": "API title",
                "description": "A very nice API",
                "version": "0.1.0",
                "contact": {
                    "name": "John Doe",
                    "email": "example@example.com",
                },
                "license_info": {
                    "name": "MIT",
                    "url": "https://opensource.org/licenses/MIT",
                },
            },
        },
    )



"""Module for audio processing."""
from mpai_cae_arp.audio import utils
from mpai_cae_arp.audio._audio import AudioWave
from mpai_cae_arp.audio._noise import Noise

__all__ = ["AudioWave", "Noise", "utils"]



"""Module containing enumerations of the standards used in the CAE-ARP standard."""
from enum import Enum


class EqualizationStandard(str, Enum):
    """An enumeration of the equalization standards used in the CAE-ARP standard.

    The following standards are available:

    - IEC
    - IEC1 (ex CCIR)
    - IEC2 (ex NAB)

    """

    IEC = "IEC"
    CCIR = "IEC1"
    NAB = "IEC2"


class SpeedStandard(float, Enum):
    """An enumeration of the tape playback speed standards used in the CAE-ARP standard.

    Options are:

    - 0.9375 ips
    - 1.875 ips
    - 3.75 ips
    - 7.5 ips
    - 15 ips
    - 30 ips

    """

    I = 0.9375  # noqa: E741
    II = 1.875
    III = 3.75
    IV = 7.5
    V = 15
    VI = 30
